/* Implement a portal engine for scenery indoor renderization.

Has the main visualization pipeline: make transformations from universe to
camera reference system, make poly clip, draw polygon mesh on frame buffer,
call recursively functions to draw other visible objects and sectors.

Portal is a polygon where other polygons are clipped, used to define visible
sectors from where the camera be.

Please refer docs/3dfc.txt for more info about the view pipeline and
engine details.

The main function is mRenderMesh. It draw a poly mesh, call:
      mDefineProjVertCoords
      mPassBackFaceCulling 
      mGetPolyTextureData
      mMakeZNearPlaneClip 

WARNING:
- polygons vertexes must be in anticlockwise order
*/



#include <string.h>

#include <math.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "texture.h"

#include "file.h"

#include "filerem.h"

#include "filepcx.h"

#include "lighttab.h"

#include "draw.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "nodever.h"

#include "nodexy.h"

#include "map.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"

#include "math.h"

#include "refsys.h"

#include "camsys.h"

#include "object.h"

#include "nodeint.h"

#include "nodetex.h"

#include "nodeobj.h"

#include "nodepor.h"

#include "nodelgt.h"

#include "filetxe.h"

#include "scenery.h"

#include "render.h"



/* constructor - init empty
**********************************************************/
CRender::CRender ()
{
   m_pastPolVertexCoords = NULL;
   m_dwMaxVertArray = 0;
   m_pastPolZClipTex = m_pastPolZClipCoords = NULL;
   m_dwMaxZClipArray = 0;

   mDestroy ();
}



/* constructor - init fly system
**********************************************************/
CRender::CRender ( CScenery& oScen, INT nResX, INT nResY, CMem& oFrBuff )
{
   m_pastPolVertexCoords = NULL;
   m_dwMaxVertArray = 0;
   m_pastPolZClipTex = m_pastPolZClipCoords = NULL;
   m_dwMaxZClipArray = 0;

   mCreate ( oScen, nResX, nResY, oFrBuff );
}



/* destructor
**********************************************************/
CRender::~CRender ()
{
   mDestroy ();
}



/* Init render engine.
oScen define the scenery to render
nResX define the X resolution to frame buffer
nResY define Y resolution to frame buffer
oFrBuffer define the frame buffer (where draw)
oCamSys the camera reference system (camera status in space)
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CRender::mCreate ( CScenery& oScen, INT nResX, INT nResY, CMem& oFrBuff )
{
   mDestroy ();

   m_poScen = &oScen;      // define scenery pointer

   // REVIEW
   // define texture vertexes, used in textures of size 256x256
   // polys with 3 vertexes
   m_astTexVertexes3 [ 0 ].nX = 0;
   m_astTexVertexes3 [ 0 ].nY = 0;
   m_astTexVertexes3 [ 1 ].nX = 0;
   m_astTexVertexes3 [ 1 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes3 [ 2 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes3 [ 2 ].nY = 0;
   // polys with 4 vertexes
   m_astTexVertexes4 [ 0 ].nX = 0;
   m_astTexVertexes4 [ 0 ].nY = 0;
   m_astTexVertexes4 [ 1 ].nX = 0;
   m_astTexVertexes4 [ 1 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 2 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 3 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 3 ].nY = 0;
   // polys with 5 vertexes
   m_astTexVertexes5 [ 0 ].nX = 0;
   m_astTexVertexes5 [ 0 ].nY = 0;
   m_astTexVertexes5 [ 1 ].nX = 0;
   m_astTexVertexes5 [ 1 ].nY = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes5 [ 2 ].nX = TEXTURE_MAXDIM / 2;
   m_astTexVertexes5 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 3 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 3 ].nY = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes5 [ 4 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 4 ].nY = 0;
   // polys with 6 vertexes
   m_astTexVertexes6 [ 0 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 1 / 4 );
   m_astTexVertexes6 [ 0 ].nY = 0;
   m_astTexVertexes6 [ 1 ].nX = 0;
   m_astTexVertexes6 [ 1 ].nY = TEXTURE_MAXDIM / 2;
   m_astTexVertexes6 [ 2 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 1 / 4 );
   m_astTexVertexes6 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 3 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes6 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 4 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 4 ].nY = TEXTURE_MAXDIM / 2;
   m_astTexVertexes6 [ 5 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes6 [ 5 ].nY = 0;
   // polys with 7 vertexes (not very precise)
   m_astTexVertexes7 [ 0 ].nX = TEXTURE_MAXDIM / 2;
   m_astTexVertexes7 [ 0 ].nY = 0;
   m_astTexVertexes7 [ 1 ].nX = 0;
   m_astTexVertexes7 [ 1 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes7 [ 2 ].nX = 0;
   m_astTexVertexes7 [ 2 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 3 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes7 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 4 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 4 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 5 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 5 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 6 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 6 ].nY = TEXTURE_MAXDIM / 3;
   // polys with 8 vertexes
   m_astTexVertexes8 [ 0 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 0 ].nY = 0;
   m_astTexVertexes8 [ 1 ].nX = 0;
   m_astTexVertexes8 [ 1 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 2 ].nX = 0;
   m_astTexVertexes8 [ 2 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 3 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 4 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 4 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 5 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 5 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 6 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 6 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 7 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 7 ].nY = 0;

   // verify Y resolution limitation
   if ( nResY > MAX_RESY )
      ABORT ( MSG_ERROR_OVER_RESY_LIMIT, __FILE__, __LINE__ )

   CLightTable // empty light table
      oLgT;
   return m_oDraw.mCreate ( oLgT, nResX, nResY, oFrBuff );
}



/* Destroy: make pointers point to NULL, free memory.
**********************************************************/
void CRender::mDestroy ()
{
   m_poScen = NULL;

   m_dwMaxVertArray = 0;
   m_dwMaxZClipArray = 0;

   // free array memory
   if ( m_pastPolVertexCoords != NULL ) {
      delete m_pastPolVertexCoords;
      m_pastPolVertexCoords = NULL;
   }

   // free array memory for clipped vertexes
   if ( m_pastPolZClipCoords != NULL ) {
      delete m_pastPolZClipCoords;
      m_pastPolZClipCoords = NULL;
   }

   // free array memory for clipped vertexes
   if ( m_pastPolZClipTex != NULL ) {
      delete m_pastPolZClipTex;
      m_pastPolZClipTex = NULL;
   }

   m_oCamVert.mRemoveAll ();
   m_oPrjVert.mRemoveAll ();

   m_oDraw.mDestroy ();
}



/* Make the scenery image in frame buffer. Call the sector
render function.
dwObjectID and dwSectorID specify in what graph node start.
poCamSys point the current camera to be used in visualization.
**********************************************************/
void CRender::mRenderFrame ( DWORD dwObjectID, DWORD dwSectorID,
                             CCameraSystem *poCamSys )
{
   // the initial clip edge buffer
   tagInterval
      astClipEdgeBuffer [ MAX_RESY ];
   m_oDraw.mResetClipEdgeBuffer ( astClipEdgeBuffer );

   // the initial clip bounding rectangle
   tagRectangle
      stClipBox;
   stClipBox.nStartX = 0;
   stClipBox.nEndX = m_oDraw.m_nMaxX;
   stClipBox.nStartY = 0;
   stClipBox.nEndY = m_oDraw.m_nMaxY;

   // clear the frame if is the case
   if ( m_poScen->m_nClearBackground != -1 )
      m_oDraw.mClear ( m_poScen->m_nClearBackground );

   // REVIEW, eh necessario?
   // define the camera system to be used in this object
   /*CCameraSystem
      oCamObj;
   // update this camera with the viewer camera, if is not the viewer
   if ( dwObjectID != m_poScen->m_dwCameraID ) {
      // set the camera as the object parameters
      oCamObj.mSetObj ( ( (CNodeObject*) m_poScen->m_oRootInstances.m_oInstances.mRetValue ( dwObjectID ) )->m_oRefSys );
      oCamObj.mTransform ( *poCamSys );
   }*/

   // make frame buffer
   mRenderSector ( dwObjectID, dwSectorID, poCamSys, 0,
                   astClipEdgeBuffer, stClipBox );
}



/* Call the sector mesh rendering.
dwObjectID define the ID of the object to be rendered
dwSectorID define the ID of the sector to be rendered
poCamSys define the camera to be used in view transformations
yLevel the actual recursion level
pastClipEdgeBuffer have the current clip buffer to be used in poly draw,
stClipBox  the clip square (if all vertexes of the poly are outside, no draw
**********************************************************/
void CRender::mRenderSector ( DWORD dwObjectID, DWORD dwSectorID,
                            CCameraSystem *poCamSys, BYTE yLevel,
                            tagInterval* pastClipEdgeBuffer,
                            tagRectangle& stClipBox )
{
   // REVIEW
   // loop bug (normal vector remove error)
   if ( yLevel > 5 )
      return;

   // what to render, parent sector or adjacents
   // define pointer to sector
   // search object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_poScen->m_oRootInstances.m_oInstances.mRetValue ( dwObjectID );
   // search sector in object
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors.mRetValue ( dwSectorID );

   // verify if sector not found
   if ( poSector == NULL )
      ABORT ( MSG_ERROR_SECTOR_NOT_FOUND, __FILE__, __LINE__ )

   // REVIEW quando nao tiver que retranformar a malha para continuar a
   // desenhar setor apos desenhar portal nao serah necessario passar
   // os IDs.
   mRenderMesh ( dwObjectID, dwSectorID, poCamSys, yLevel,
                 pastClipEdgeBuffer, stClipBox,
                 poSector->m_oVertexes, poSector->m_oPolygons,
                 poSector->m_oPortals );

   // search for objects to render (objects inside this sector)
   DWORD
      dwNumObjects = poSector->m_oInstances.mRetSize ();
   // for each object in sector
   for ( DWORD dwcO = 0; dwcO < dwNumObjects; dwcO ++ ) {
      
      // the object ID
      DWORD
         dwObjID = poSector->m_oInstances.mRetKey ( dwcO );
      // the object inside this sector
      CNodeObject
         *poObj = (CNodeObject*) poSector->m_oInstances.mRetValue ( dwObjID );

      // define the camera system to be used in this object
      CCameraSystem
         oCamObj;
      // update this camera with the viewer camera, if is not the viewer
      if ( dwObjID != m_poScen->m_dwCameraID ) {
         // set the camera as the object parameters
         oCamObj.mSetObj ( poObj->m_oRefSys );
         oCamObj.mTransform ( *poCamSys );
      }
      
      // REVIEW nao usar dwObjectID, dwSectorID
      mRenderMesh ( 1000, 1000, &oCamObj, yLevel,
                    pastClipEdgeBuffer, stClipBox,
                    poObj->m_oVertexes, poObj->m_oPolygons,
                    poObj->m_oPortals );
   }
}



/* Draw a poly mesh: transform to SRC, project and
ajust to frame buffer coordenates, draw polygons.
If a portal is visible call recursively the draw function for
the respective sector or object.
**********************************************************/
void CRender::mRenderMesh ( DWORD dwObjectID, DWORD dwSectorID,
                            CCameraSystem *poCamSys, BYTE yLevel,
                            tagInterval* pastClipEdgeBuffer,
                            tagRectangle& stClipBox,
                            CList& oMeshVertexes, CList& oMeshPolygons,
                            CMap& oMeshPortals )
{
   // transform sector mesh from universe to camera
   mApplyCamera ( poCamSys, oMeshVertexes );

   // project mesh to 2D and ajust to viewport reference system
   mAjust ( poCamSys );

   // draw poly mesh making portal clip
   DWORD
      dwPolyNdx = 0;  // the index of current poly beeing draw
   // for each polygon in mesh
   oMeshPolygons.mSelectStart ();
   while ( oMeshPolygons.mRetExistSelection () ) {

      // REVIEW
      if ( dwPolyNdx > 300 )
         ABORT ( "Estourou numero maximo de poligonos (bug)\n",__FILE__,__LINE__)

      // the current polygon to define
      CNodePolygon
         *poPol = (CNodePolygon*) oMeshPolygons.mRetSelected ();

      // define poly vertex coordenates projected, make frame buffer clip
      // if some vertex is inside the frame buffer -> draw poly, else no
      DWORD
         dwBehind;    // count number of vertexes behind near Z clip plane
      if ( mDefineProjVertCoords ( poPol, stClipBox, dwBehind ) )

         // make normal vector visible cut
         if ( mPassBackFaceCulling () ) {

            // define polygon texture and texture coodenates
            CTexture
               *poTx = NULL;
            tagXY
               *pastTexVertexes = NULL;
            mGetPolyTextureData ( poPol, &poTx, &pastTexVertexes ); 

            // point the poly data to be drawn
            tagXY
               *pstPolVerDraw = NULL,
               *pstPolTexDraw = NULL;
            DWORD
               dwPolNumVerDraw;

            // define if the polygon is clipped in Z plan
            // verify if poly not intercect the Z plan
            if ( dwBehind == 0 ) {

               // define the polygon to draw as the current polygon
               pstPolVerDraw = m_pastPolVertexCoords;
               dwPolNumVerDraw = poPol->m_dwNumVertexes;
               // if texture defined
               if ( poTx != NULL )
                  pstPolTexDraw = pastTexVertexes;
            }
            else{ // if the poly intersect the clip Z plan make poly cut

               dwPolNumVerDraw = mMakeZNearPlaneClip ( poPol, pastTexVertexes, poCamSys, pastClipEdgeBuffer, stClipBox );

               // define the polygon to draw as the current polygon
               pstPolVerDraw = m_pastPolZClipCoords;
               // if texture defined
               if ( poTx != NULL )
                  pstPolTexDraw = m_pastPolZClipTex;
            }

            // verify if the polygon is a portal
            if ( poPol->m_bPortal == TRUE ) {

               // if yes make the new clip edge buffer
               // have start and end X limits per scanline (the edge buffer)
               tagInterval
                  astNewBuffer [ MAX_RESY ];
               // copy the current buffer to new (will be used to define new)
               memcpy ( (void*) astNewBuffer, (void*) pastClipEdgeBuffer, sizeof ( tagInterval ) * m_oDraw.m_nResY );
               // define new clip box (rectangle around clip poly)
               tagRectangle
                  stNewBox;
               stNewBox.nStartX = stClipBox.nStartX;
               stNewBox.nStartY = stClipBox.nStartY;
               stNewBox.nEndX = stClipBox.nEndX;
               stNewBox.nEndY = stClipBox.nEndY;

               // verify if need draw a new sector
               // Set the new clip edge buffer using the clip polygon (portal)
               // if the clip polygon is totaly clipped (return FALSE) the destin
               // sector isnt visible and to not call sector draw, recursion end
               if ( m_oDraw.mSetClipEdgeBuffer ( pstPolVerDraw, dwPolNumVerDraw, astNewBuffer, stNewBox ) ) {

                  // the sector is visible
                  // define the object and sector IDs
                  CNodePortal
                     *poDestin = (CNodePortal*) oMeshPortals.mRetValue ( dwPolyNdx );
                  // call draw again
                  mRenderSector ( poDestin->m_stPortal.dwObjectID, poDestin->m_stPortal.dwSectorID, poCamSys, yLevel + 1, astNewBuffer, stNewBox );

                  // REVIEW (guardar numa "cache")
                  // reselect object, sector and polygon
                  CNodeObject
                     *poObj = (CNodeObject*) m_poScen->m_oRootInstances.m_oInstances.mRetValue ( dwObjectID );
                  // verify if object not found
                  if ( poObj == NULL )
                     ABORT ( MSG_ERROR_OBJECT_NOT_FOUND, __FILE__, __LINE__ )
                  // reselect
                  CNodeSector
                     *poSector = (CNodeSector*) poObj->m_oSectors.mRetValue ( dwSectorID );
                  poSector->m_oPolygons.mSelectNode ( dwPolyNdx );
                  // transform sector mesh from universe to camera
                  mApplyCamera ( poCamSys, poSector->m_oVertexes );
                  // project mesh to 2D
                  mAjust ( poCamSys );
               }
            }
            else // if poly is not a portal draw texture (bug: "global" projected
                 // member variables are changes after other sectors draw->
                 // impossible draw a textured portal now)
               // draw textured polygon
               // if texture defined
               if ( poTx != NULL ) {

                  // set the light table to be used in poly draw
                  CNodeLightTable
                     *poNodeLgT = (CNodeLightTable*) m_poScen->m_oLightTables.mRetValue ( poPol->m_dwLightID );
                  // verify if not found
                  if ( poNodeLgT == NULL )
                     ABORT ( MSG_ERROR_NO_LIGHTTABLE, __FILE__, __LINE__ )
                  m_oDraw.m_poLightTable = &poNodeLgT->m_oLgT;

                  // draw it
                  m_oDraw.mTexturedPolygon ( pstPolVerDraw, pstPolTexDraw, dwPolNumVerDraw,
                                             *poTx, pastClipEdgeBuffer, poPol->m_fLight );
                  //m_oDraw.mFlatPolygon ( pstPolVerDraw, dwPolNumVerDraw, 168+poPol->m_dwTextureID );
               }
         }

      // next polygon to draw
      oMeshPolygons.mSelectNext ();
      dwPolyNdx ++;
   }
}



/* Make universe reference system to view reference system 
transformation. Apply the reference system base and origin in oSrc
to CRender::m_oCamVert
The given list must have CNodeVertex nodes.
poCamSys is the camera to be used in transformation.
NOTE: the fix point base components of camera system MUST
be updated.
**********************************************************/
void CRender::mApplyCamera ( CCameraSystem *poCamSys, CList& oSrc )
{
   DWORD
      dwNumVert = oSrc.mRetNumNodes ();  // number of nodes

   // verify if need realocate memory for destin list (not enough nodes)
   if ( dwNumVert > m_oCamVert.mRetNumNodes () ) {

      // disable direct access feature
      m_oCamVert.mDisableDirectAccess ();

      // define how many nodes must be added
      DWORD
         dwNew = dwNumVert - m_oCamVert.mRetNumNodes ();
      // alocate more nodes for destin
      for ( DWORD dwcN = 0; dwcN < dwNew; dwcN ++ ) {

         CNodeVertex
            *poNew = new CNodeVertex;
         // verify alocation error
         if ( poNew == NULL )
            ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

         m_oCamVert.mSetEnd ( (CNode*) poNew );
      }

      // enable direct access feature
      m_oCamVert.mEnableDirectAccess ();
   }

   // transform vertexes (move and rotate)
   tagXYZ
      stPos;
   stPos.nX = (INT) poCamSys->m_stPos.fX;
   stPos.nY = (INT) poCamSys->m_stPos.fY;
   stPos.nZ = (INT) poCamSys->m_stPos.fZ;
   // select first vertex of lists
   oSrc.mSelectStart ();
   m_oCamVert.mSelectStart ();

   // scan all source list transforming to destin list
   while ( oSrc.mRetExistSelection () ) {

      // point vertexes
      CNodeVertex
         *poVSrc = (CNodeVertex*) oSrc.mRetSelected ();
      CNodeVertex
         *poVDest = (CNodeVertex*) m_oCamVert.mRetSelected ();
      // apply reference system origin (move)

      poVDest->m_stVertex.nX = poVSrc->m_stVertex.nX - stPos.nX;
      poVDest->m_stVertex.nY = poVSrc->m_stVertex.nY - stPos.nY;
      poVDest->m_stVertex.nZ = poVSrc->m_stVertex.nZ - stPos.nZ;// REVIEW - m_poCamSys->m_nDistCP;

      // apply reference system base (rotate)
      tagXYZ
         stRot;   // temp result
      stRot.nX = ( poVDest->m_stVertex.nX * poCamSys->m_stFixI.nX >> 16 ) +
                 ( poVDest->m_stVertex.nY * poCamSys->m_stFixI.nY >> 16 ) +
                 ( poVDest->m_stVertex.nZ * poCamSys->m_stFixI.nZ >> 16 );
      stRot.nY = ( poVDest->m_stVertex.nX * poCamSys->m_stFixJ.nX >> 16 ) +
                 ( poVDest->m_stVertex.nY * poCamSys->m_stFixJ.nY >> 16 ) +
                 ( poVDest->m_stVertex.nZ * poCamSys->m_stFixJ.nZ >> 16 );
      stRot.nZ = ( poVDest->m_stVertex.nX * poCamSys->m_stFixK.nX >> 16 ) +
                 ( poVDest->m_stVertex.nY * poCamSys->m_stFixK.nY >> 16 ) +
                 ( poVDest->m_stVertex.nZ * poCamSys->m_stFixK.nZ >> 16 );
      poVDest->m_stVertex.nX = stRot.nX;
      poVDest->m_stVertex.nY = stRot.nY;
      poVDest->m_stVertex.nZ = stRot.nZ;// + m_poCamSys->m_nDistCP;

      // next node
      oSrc.mSelectNext ();
      m_oCamVert.mSelectNext ();
   }
}



/* Make de projection from a 3D mesh to a 2D mesh and ajust
to viewport. Use m_oCamVert as input and m_oPrjVert as output.
poCamSys is the camera to be used in transformation.
**********************************************************/
void CRender::mAjust ( CCameraSystem *poCamSys )
{
   // verify if need grow projected vertexes list (insert new nodes)
   if ( m_oPrjVert.mRetNumNodes () < m_oCamVert.mRetNumNodes () ) {

      // number of nodes to insert
      DWORD
         dwNumNew = m_oCamVert.mRetNumNodes () - m_oPrjVert.mRetNumNodes ();

      // no direct access (need to allow list pointers change)
      m_oPrjVert.mDisableDirectAccess ();

      // add vertexes
      for ( DWORD dwcD = 0; dwcD < dwNumNew; dwcD ++ ) {
         CNodeXY
            *poNode = new CNodeXY;
         if ( poNode == NULL )
            ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
         // add to vertex list
         m_oPrjVert.mSetEnd ( poNode );
      }

      // direct access (for fast random access)
      m_oPrjVert.mEnableDirectAccess ();
   }

   // for each vertex
   m_oCamVert.mSelectStart ();
   m_oPrjVert.mSelectStart ();
   while ( m_oCamVert.mRetExistSelection () ) {

      // vertex to project
      CNodeVertex
         *poVert = (CNodeVertex*) m_oCamVert.mRetSelected ();
      // projected vertex
      CNodeXY
         *poXY = (CNodeXY*) m_oPrjVert.mRetSelected ();

      // project
      // verify if vertex is in front of the projection plane
      if ( poVert->m_stVertex.nZ > poCamSys->m_nDistCP ) {

         // make standard projection
         poXY->m_stXY.nX = (INT) ( (float) poVert->m_stVertex.nX / ( (float) poVert->m_stVertex.nZ / - poCamSys->m_nDistCP + 1 ) );
         poXY->m_stXY.nY = (INT) ( (float) poVert->m_stVertex.nY / ( (float) poVert->m_stVertex.nZ / - poCamSys->m_nDistCP + 1 ) );
      }
      else {
         // REVIEW - soh projeta se positivo, senao nem aloca memoria,
         //     seta flag no pol...? ou aloca e nao usa, seta flag X=-1
         // make interpolated projection
         poXY->m_stXY.nX = (INT) ( (float) poVert->m_stVertex.nX / ( (float) poVert->m_stVertex.nZ / (float) ( 1 - poVert->m_stVertex.nZ ) + 1 ) );
         poXY->m_stXY.nY = (INT) ( (float) poVert->m_stVertex.nY / ( (float) poVert->m_stVertex.nZ / (float) ( 1 - poVert->m_stVertex.nZ ) + 1 ) );
      }

      // ajust for viewport (middle frame buffer, invert Y, scale)
      poXY->m_stXY.nX = poXY->m_stXY.nX + m_oDraw.m_nMidX;
      poXY->m_stXY.nY = m_oDraw.m_nResY - ( poXY->m_stXY.nY + m_oDraw.m_nMidY );

      // next vertex
      m_oCamVert.mSelectNext ();
      m_oPrjVert.mSelectNext ();
   }
}



/* Define polygon vertexes projected coordenates and make frame buffer clip.
   For clipping is counted the number of vertexes that are outside
   the clip box area. Is a first clip method, not realy powerfull.
Update: m_dwMaxVertArray, m_pastPolVertexCoords
return TRUE - poly is totaly or partialy visible / FALSE - invisible
return dwBehind - number of vertexes behind the near Z clip plane
**********************************************************/
BOOL CRender::mDefineProjVertCoords ( CNodePolygon *poPol,
                                      tagRectangle& stClipBox,
                                      DWORD &dwBehind )
{
   // counters for the number of vertexes outside the clip area
   DWORD
      dwLeft = 0,
      dwRight = 0,
      dwTop = 0,
      dwBottom = 0;
   dwBehind = 0;
   
   // verify if have enough space in poly vertex array, if no realocate
   if ( m_dwMaxVertArray < poPol->m_dwNumVertexes ) {

      // realocate memory
      // desalocate
      delete m_pastPolVertexCoords;
      // alocate
      if ( ( m_pastPolVertexCoords = new tagXY [ poPol->m_dwNumVertexes ] ) == NULL )
         ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
      m_dwMaxVertArray = poPol->m_dwNumVertexes;
   }

   // scan polygon vertex indexes and copy respective coordenates to array
   DWORD
      *pdwNdxVert;        // point vertex indexes array in polygon
   pdwNdxVert = poPol->m_padwVertexes;
   tagXY
      *pstPolVertexes;    // point vertex coordenate array
   pstPolVertexes = m_pastPolVertexCoords;
   for ( DWORD dwcI = 0; dwcI < poPol->m_dwNumVertexes; dwcI ++ ) {

      // define vertex projected coordenate
      CNodeXY
         *poNodeVert = (CNodeXY*) m_oPrjVert.mRetNode ( *pdwNdxVert );

      // define the 3D camera vertex coordenate
      CNodeVertex
         *poNodeVertCam = (CNodeVertex*) m_oCamVert.mRetNode ( *pdwNdxVert );

      // copy coords
      pstPolVertexes->nX = poNodeVert->m_stXY.nX;
      pstPolVertexes->nY = poNodeVert->m_stXY.nY;

      // verify if vertex is outside visible area

      // verify if is behind the clip plan
      if ( poNodeVertCam->m_stVertex.nZ < Z_CLIP )
         dwBehind ++;

      // verify if is outside frame buffer: left, right
      if ( pstPolVertexes->nX < stClipBox.nStartX )
         dwLeft ++;
      else
         if ( pstPolVertexes->nX > stClipBox.nEndX )
            dwRight ++;

      // verify top and bottom
      if ( pstPolVertexes->nY < stClipBox.nStartY )
         dwTop ++;
      else
         if ( pstPolVertexes->nY > stClipBox.nEndY )
            dwBottom ++;

      // next arrays positions (test next)
      pdwNdxVert ++;
      pstPolVertexes ++;
   }

   // verify if all vertexes are outside the frame buffer, if no -> draw
   if ( dwLeft   != poPol->m_dwNumVertexes &&
        dwRight  != poPol->m_dwNumVertexes &&
        dwTop    != poPol->m_dwNumVertexes &&
        dwBottom != poPol->m_dwNumVertexes &&
        dwBehind != poPol->m_dwNumVertexes )
      return TRUE;

   // no poly draw
   return FALSE;
}
          

          
/* Perform back-face culling: verify if visible poly face is in front
of the view plane. For this is compared the poly normal vector Z value.
return TRUE - poly visible / FALSE - poly invisible
**********************************************************/
BOOL CRender::mPassBackFaceCulling ()
{
   // REVIEW - now does not work very well: when you is close->overflow
   /* fazer ateh dot-product e olhar sinal. usar vetores no SRC
   vetor de ti ate vertice mais proximo em Z?
   Parece que o problema nao eh soh overflow......
   // make normal vector visible cut
   tagfXYZ
      stNormalCam;
   // polygon normal in SRU
   stNormalCam.fX = poPol->m_stNormal.fX;
   stNormalCam.fY = poPol->m_stNormal.fY;
   stNormalCam.fZ = poPol->m_stNormal.fZ;
   // polygon normal in SRC
   poCamSys->mApplyBase ( stNormalCam );
   */

   // define 3 vertexes inside polygon plan, use project x y
   tagXY
      *pstVert0 = &m_pastPolVertexCoords [ 0 ],
      *pstVert1 = &m_pastPolVertexCoords [ 1 ],
      *pstVert2 = &m_pastPolVertexCoords [ 2 ];

   // define 2 vectors (x,y) components inside the poly plan
   tagfXYZ
      stXY1,
      stXY2;
   stXY1.fX = (float) ( pstVert0->nX - pstVert1->nX ) / 1000;
   stXY1.fY = (float) ( pstVert0->nY - pstVert1->nY ) / 1000;
   stXY2.fX = (float) ( pstVert0->nX - pstVert2->nX ) / 1000;
   stXY2.fY = (float) ( pstVert0->nY - pstVert2->nY ) / 1000;

   // define only the normal Z value
   float
      fNormalZ = stXY1.fX * stXY2.fY - stXY1.fY * stXY2.fX;

   // verify if is negative, if yes draw poly (poly visible)
   if ( fNormalZ < 0 )
      return TRUE;

   return FALSE;
}



/* Get texture bitmap and texel coordenates for a polygon.
If the texture ID == -1 -> poly with no texture
return ppoTx - point the bitmap pointer (bitmap pointer get NULL if no texture)
       ppastTexVertexes - point the texels coord pointer
                          (pointed pointer is NULL if no texture)
       Texels are texture coodenates.
**********************************************************/
void CRender::mGetPolyTextureData ( CNodePolygon *poPol,
                                    CTexture **ppoTx,
                                    tagXY **ppastTexVertexes )
{
   // define polygon texture
   // point texture or NULL if no texture defined
   CNodeTexture
      *poNodeTx = (CNodeTexture*) m_poScen->m_oTextures.mRetValue ( poPol->m_dwTextureID );

   // verify if texture not found, or poly with no texture (ID == -1)
   if ( poNodeTx == NULL )
      // verify if texture ID == -1 -> no texture for this poly, else error
      if ( (INT) poPol->m_dwTextureID != -1 )
         ABORT ( MSG_ERROR_NO_TEXTURE, __FILE__, __LINE__ )

   // define texture coordenates
   // if texture defined
   if ( poNodeTx != NULL ) {

      // point texture object
      *ppoTx = &poNodeTx->m_oTx;

      // select a predefined array of texture vertexes
      switch ( poPol->m_dwNumVertexes ) {

         case 3:
            *ppastTexVertexes = m_astTexVertexes3;
            break;

         case 4:
            *ppastTexVertexes = m_astTexVertexes4;
            break;

         case 5:
            *ppastTexVertexes = m_astTexVertexes5;
            break;

         case 6:
            *ppastTexVertexes = m_astTexVertexes6;
            break;

         case 7:
            *ppastTexVertexes = m_astTexVertexes7;
            break;

         case 8:
            *ppastTexVertexes = m_astTexVertexes8;
            break;

         default:
            ABORT ( MSG_ERROR_MANY_VERTEXES, __FILE__, __LINE__ )
            break;
      }
   }
   else {
      // no texture defined
      *ppoTx = NULL;
      *ppastTexVertexes = NULL;
   }
}
   


/* Create a new poly clipped with the near Z plane.
Update: m_pastPolZClipCoords, m_pastPolZClipTex, m_dwMaxZClipArray
with the new poly data.
return the number of vertexes in new poly.
**********************************************************/
DWORD CRender::mMakeZNearPlaneClip ( CNodePolygon* poPol,
                                     tagXY *pastTexVertexes,
                                     CCameraSystem *poCamSys,
                                     tagInterval* pastClipEdgeBuffer,
                                     tagRectangle& stClipBox )
{ 
   // define number of vertexes in polygon
   DWORD
      dwNumVertexes = poPol->m_dwNumVertexes;

   // search one point in the visible side of Z plan
   DWORD
      *pdwNdxVertClip;    // point vertex indexes array in poly
   pdwNdxVertClip = poPol->m_padwVertexes;
   DWORD
      dwNdxFirst = 0;     // the index of the vertex found
   // scan all polygon vertexes
   for ( DWORD dwcV = 0; dwcV < dwNumVertexes; dwcV ++ )
      // verify if vertex is in front of the Z plan
      if ( ( (CNodeVertex*) m_oCamVert.mRetNode ( *pdwNdxVertClip ) )->m_stVertex.nZ >= Z_CLIP )
         break;
      else {
         pdwNdxVertClip ++;
         dwNdxFirst ++;
      }

   // verify if need grow the vertexes array in clipped poly
   if ( m_dwMaxZClipArray < ( m_dwMaxVertArray + 1 ) ) {
      // +1 for new clipped vertex!

      // realocate memory for polygon vertexes
      // desalocate
      if ( m_pastPolZClipCoords != NULL )
         delete m_pastPolZClipCoords;
      // alocate
      if ( ( m_pastPolZClipCoords = new tagXY [ dwNumVertexes + 1 ] ) == NULL )
         ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

      // if texture defined
      if ( pastTexVertexes != NULL ) {
         // realocate memory for polygon texture vertexes
         // desalocate
         if ( m_pastPolZClipTex != NULL )
            delete m_pastPolZClipTex;
         // alocate
         if ( ( m_pastPolZClipTex = new tagXY [ dwNumVertexes + 1 ] ) == NULL )
            ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )
      }

      // define new number of vertexes
      m_dwMaxZClipArray = dwNumVertexes + 1;
   }

   // if the polygon is a portal, reset the clip edge buffer and set
   // poly to frame buffer size. Don't continue the rest of clip.
   if ( poPol->m_bPortal == TRUE ) {

      // set new poly data
      // coords
      m_pastPolZClipCoords [ 0 ].nX = 0;
      m_pastPolZClipCoords [ 0 ].nY = 0;
      m_pastPolZClipCoords [ 1 ].nX = 0;
      m_pastPolZClipCoords [ 1 ].nY = m_oDraw.m_nMaxY;
      m_pastPolZClipCoords [ 2 ].nX = m_oDraw.m_nMaxX;
      m_pastPolZClipCoords [ 2 ].nY = m_oDraw.m_nMaxY;
      m_pastPolZClipCoords [ 3 ].nX = m_oDraw.m_nMaxX;
      m_pastPolZClipCoords [ 3 ].nY = 0;
      // texels
      memcpy ( (void*) m_pastPolZClipTex, (void*) m_astTexVertexes4, sizeof ( tagXY ) * 4 );
      
      // set clip edge buffer
      m_oDraw.mResetClipEdgeBuffer ( pastClipEdgeBuffer );
      // the initial clip bounding rectangle
      stClipBox.nStartX = 0;
      stClipBox.nEndX = m_oDraw.m_nMaxX;
      stClipBox.nStartY = 0;
      stClipBox.nEndY = m_oDraw.m_nMaxY;

      return 4;
   }

   // copy vertexes while they are in front of the Z clip plan

   tagXY
      // source coord
      *pstSrcCoord = &m_pastPolVertexCoords [ dwNdxFirst ],
      // destin coord
      *pstDestCoord = m_pastPolZClipCoords,
      // texture coords
      *pstSrcTex = NULL,
      *pstDestTex = NULL,
      // the end of vertex coord array
      *pstEndPolVertexCoords = &m_pastPolVertexCoords [ dwNumVertexes - 1 ];
   // if texture defined
   if ( pastTexVertexes != NULL ) {
      // src texture coord
      pstSrcTex = &pastTexVertexes [ dwNdxFirst ];
      // dest texture coord
      pstDestTex = m_pastPolZClipTex;
   }
   DWORD
      // the index of source vertex (in polygon array)
      dwNdxSrc = dwNdxFirst,
      // the index of last copyed vertex (in polygon array)
      dwNdxLastSrc,
      // number of vertexes in new poly
      dwNumVert = 0;
   DWORD
      // point poly vertex index array
      *pdwVertNdxSrc = &poPol->m_padwVertexes [ dwNdxFirst ];
   // scan, for each vertex
   do {

      // copy
      pstDestCoord->nX = pstSrcCoord->nX;
      pstDestCoord->nY = pstSrcCoord->nY;
      // if texture defined
      if ( pastTexVertexes != NULL ) {
         pstDestTex->nX = pstSrcTex->nX;
         pstDestTex->nY = pstSrcTex->nY;
      }
      // define last data copyed
      dwNdxLastSrc = dwNdxSrc;

      // define next poly vertex and texture vertex to copy
      // verify if is in the end of poly vertex array
      if ( pstSrcCoord == pstEndPolVertexCoords ) {
         // goto start
         pstSrcCoord = m_pastPolVertexCoords;
         pstDestCoord ++;
         dwNdxSrc = 0;
         pdwVertNdxSrc = poPol->m_padwVertexes;
         dwNumVert ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstSrcTex = pastTexVertexes;
            pstDestTex ++;
         }
      }
      else {
         // REVIEW - esse monte de ++ nao eh mais lento que 1 ndx?
         // set to next position
         pstSrcCoord ++;
         pstDestCoord ++;
         dwNdxSrc ++;
         pdwVertNdxSrc ++;
         dwNumVert ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstDestTex ++;
            pstSrcTex ++;
         }
      }
   } while ( ( (CNodeVertex*) m_oCamVert.mRetNode ( *pdwVertNdxSrc ) )->m_stVertex.nZ >= Z_CLIP );

   // define the intercection point between the line and the plan
   // the 2 point that compose the line
   CNodeVertex
      *poVert1 = (CNodeVertex*) m_oCamVert.mRetNode ( poPol->m_padwVertexes [ dwNdxSrc ] ),
      *poVert2 = (CNodeVertex*) m_oCamVert.mRetNode ( poPol->m_padwVertexes [ dwNdxLastSrc ] );
   tagXY
      // texture points
      *pstTex1,
      *pstTex2;
   float
      // define delta Z rate between the points
      fZRate = ( (float) Z_CLIP - poVert1->m_stVertex.nZ ) / ( poVert2->m_stVertex.nZ - poVert1->m_stVertex.nZ );
   // define new polygon x and y in camera reference system
   pstDestCoord->nX = (INT) ( poVert1->m_stVertex.nX + ( ( poVert2->m_stVertex.nX - poVert1->m_stVertex.nX ) * fZRate ) );
   pstDestCoord->nY = (INT) ( poVert1->m_stVertex.nY + ( ( poVert2->m_stVertex.nY - poVert1->m_stVertex.nY ) * fZRate ) );
   // project vertex
   pstDestCoord->nX = (INT) ( (float) pstDestCoord->nX / ( (float) Z_CLIP / - poCamSys->m_nDistCP + 1 ) );
   pstDestCoord->nY = (INT) ( (float) pstDestCoord->nY / ( (float) Z_CLIP / - poCamSys->m_nDistCP + 1 ) );
   // just to view port
   pstDestCoord->nX = pstDestCoord->nX + m_oDraw.m_nMidX;
   pstDestCoord->nY = m_oDraw.m_nResY - ( pstDestCoord->nY + m_oDraw.m_nMidY );
   // if texture defined
   if ( pastTexVertexes != NULL ) {
      // first texture point (behind clip plan)
      pstTex1 = &pastTexVertexes [ dwNdxSrc ],
      // second texture point (in front of clip plan)
      pstTex2 = &pastTexVertexes [ dwNdxLastSrc ];
      // define new texture x and y (use float for better precision)
      pstDestTex->nX = pstTex1->nX + (INT) ( (float) ( pstTex2->nX - pstTex1->nX ) * fZRate );
      pstDestTex->nY = pstTex1->nY + (INT) ( (float) ( pstTex2->nY - pstTex1->nY ) * fZRate );
      pstDestTex ++;
   }
   pstDestCoord ++;
   dwNumVert ++;

   // skip vertexes that are behind the plan
   do {
      // define last data copyed
      dwNdxLastSrc = dwNdxSrc;
      // define next poly vertex and texture vertex to copy
      // verify if is in the end of poly vertex array
      if ( pstSrcCoord == pstEndPolVertexCoords ) {
         // goto start
         pstSrcCoord = m_pastPolVertexCoords;
         dwNdxSrc = 0;
         pdwVertNdxSrc = poPol->m_padwVertexes;
         // if texture defined
         if ( pastTexVertexes != NULL )
            pstSrcTex = pastTexVertexes;
      }
      else {
         // set to next position
         pstSrcCoord ++;
         dwNdxSrc ++;
         pdwVertNdxSrc ++;
         // if texture defined
         if ( pastTexVertexes != NULL )
            pstSrcTex ++;
      }
   } while ( ( (CNodeVertex*) m_oCamVert.mRetNode ( *pdwVertNdxSrc ) )->m_stVertex.nZ < Z_CLIP );

   // define the intercection point between the line and the plan
   // the 2 point that compose the line
   poVert2 = (CNodeVertex*) m_oCamVert.mRetNode ( poPol->m_padwVertexes [ dwNdxSrc ] ),
   poVert1 = (CNodeVertex*) m_oCamVert.mRetNode ( poPol->m_padwVertexes [ dwNdxLastSrc ] );
   // define delta Z rate between the points
   fZRate = ( (float) Z_CLIP - poVert1->m_stVertex.nZ ) / ( poVert2->m_stVertex.nZ - poVert1->m_stVertex.nZ );
   // define new polygon x and y in camera reference system
   pstDestCoord->nX = (INT) ( poVert1->m_stVertex.nX + ( ( poVert2->m_stVertex.nX - poVert1->m_stVertex.nX ) * fZRate ) );
   pstDestCoord->nY = (INT) ( poVert1->m_stVertex.nY + ( ( poVert2->m_stVertex.nY - poVert1->m_stVertex.nY ) * fZRate ) );
   // project vertex
   pstDestCoord->nX = (INT) ( (float) pstDestCoord->nX / ( (float) Z_CLIP / - poCamSys->m_nDistCP + 1 ) );
   pstDestCoord->nY = (INT) ( (float) pstDestCoord->nY / ( (float) Z_CLIP / - poCamSys->m_nDistCP + 1 ) );
   // just to view port
   pstDestCoord->nX = pstDestCoord->nX + m_oDraw.m_nMidX;
   pstDestCoord->nY = m_oDraw.m_nResY - ( pstDestCoord->nY + m_oDraw.m_nMidY );
   // if texture defined
   if ( pastTexVertexes != NULL ) {
      // first texture point (behind clip plan)
      pstTex2 = &pastTexVertexes [ dwNdxSrc ],
      // second texture point (in front of clip plan)
      pstTex1 = &pastTexVertexes [ dwNdxLastSrc ];
      // define new texture x and y (use float for better precision)
      pstDestTex->nX = pstTex1->nX + (INT) ( (float) ( pstTex2->nX - pstTex1->nX ) * fZRate );
      pstDestTex->nY = pstTex1->nY + (INT) ( (float) ( pstTex2->nY - pstTex1->nY ) * fZRate );
      pstDestTex ++;
   }
   pstDestCoord ++;
   dwNumVert ++;

   // copy rest of vertexes while be found the start vertex
   while ( dwNdxSrc != dwNdxFirst ) {
      // copy
      pstDestCoord->nX = pstSrcCoord->nX;
      pstDestCoord->nY = pstSrcCoord->nY;
      // if texture defined
      if ( pastTexVertexes != NULL ) {
         pstDestTex->nX = pstSrcTex->nX;
         pstDestTex->nY = pstSrcTex->nY;
      }
      // define next poly vertex and texture vertex to copy
      // verify if is in the end of poly vertex array
      if ( pstSrcCoord == pstEndPolVertexCoords ) {
         // goto start
         pstSrcCoord = m_pastPolVertexCoords;
         pstDestCoord ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstSrcTex = pastTexVertexes;
            pstDestTex ++;
         }
         dwNdxSrc = 0;
         pdwVertNdxSrc = poPol->m_padwVertexes;
         dwNumVert ++;
      }
      else {
         // set to next position
         pstSrcCoord ++;
         pstDestCoord ++;
         dwNdxSrc ++;
         pdwVertNdxSrc ++;
         dwNumVert ++;
         // if texture defined
         if ( pastTexVertexes != NULL ) {
            pstSrcTex ++;
            pstDestTex ++;
         }
      }
   }

   return dwNumVert;
}
